#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <omp.h>
#include <time.h>

/* Compilation : icc -O2 -openmp -o exec main-2d-openmp.c */

main(void)
{
  /* Dclarations */
  int n=1000000, kx=512, ky=512;
  double H[512][512];

  int i, j, l;
  double X[512], Y[512], res[512][512], R[1000000], thetaU[1000000], thetaZ[1000000], tmp1, tmp2, tmp3, tmp, pasX, pasY;
  FILE * fic;
  double s;

  /* Initializations */

  pasX = 0.01;
  pasY = 0.01;

  X[0]=pasX;
  Y[0]=pasY;
  for (i=1;i<kx;i++)
    {
      X[i]=X[i-1]+pasX;
      Y[i]=Y[i-1]+pasY;
    }

  for (i=0;i<kx;i++) for (j=0;j<ky;j++)
    {
      res[i][j]=0;
      H[i][j]=0.2;
    }

  /* Random generation */
  srand(time(NULL));
  R[0]= -log(rand()/(float)RAND_MAX);

  for (l=0;l<n;l++)
    {
      thetaU[l] = rand()/(float)RAND_MAX; /* Uniform distribution in [0,1] */
      thetaZ[l] = rand()/(float)RAND_MAX; /* Uniform distribution in  [0,1] */
      if (l!=n-1) R[l+1] = R[l] - log(rand()/(float)RAND_MAX); /* Cumulated sum of n observations from exponential distribution */
    }

  /* End of initializations */

#pragma omp parallel private(tmp,tmp1,tmp2,tmp3,l) 
 {
  #pragma omp for
  for (i=0;i<kx;i++)
  #pragma omp parallel shared(i)
    {
    #pragma omp for
    for (j=0;j<ky;j++)
      {
	/* Computation of the process value for the cell [i][j] */
        for (l=0;l<n;l++)
          {
            /* Intermediate computation */
            tmp1 = sqrt(R[l]/M_PI)*(X[i]*cos(2*M_PI*thetaU[l])+Y[j]*sin(2*M_PI*thetaU[l]));
            tmp2 = (cos(tmp1)-1)*cos(2*M_PI*thetaZ[l])+sin(tmp1)*sin(2*M_PI*thetaZ[l]);
            tmp3 = pow(R[l],(1+H[i][j])/2);
            tmp = tmp2/tmp3;
	    res[i][j] = res[i][j] + tmp;
          } /* End of loop l */
        /* Computation of the final term for the cell [i][j] */
        res[i][j] = res[i][j] * 2*pow(M_PI,0.5*(1+H[i][j]));

      } /* End of loop j */
    } /* End of loop i */
 } /* End of parallel area */

  /* Results are written into a file */
      fic = fopen("res-simul.out","w");
      for (i=0;i<kx;i++)
        {
          for (j=0;j<ky;j++) fprintf(fic,"%.3f ",res[i][j]);
          fprintf(fic,"\n");
        }
      fclose(fic);
}
